/*
 * Copyright (C) 2019-2020  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/proto_wrap.hh"

#include "liblogger/log_entry.hh"
#include "liblogger/log_entry_ident.hh"
#include "liblogger/session_ident.hh"

#include <boost/asio/ip/address.hpp>
#include <boost/date_time/posix_time/posix_time.hpp>

#include <map>
#include <string>
#include <sstream>

namespace Fred {
namespace Logger {

namespace {

void proto_wrap(const boost::posix_time::ptime& _src, google::protobuf::Timestamp* _dst)
{
    const boost::posix_time::ptime epoch(boost::gregorian::date(1970, boost::date_time::Jan, 1),
            boost::posix_time::time_duration(0, 0, 0));
    const boost::posix_time::time_duration diff = _src - epoch;
    long long total_seconds = diff.total_seconds();
    long long fractional_seconds = diff.total_nanoseconds() % 1'000'000'000;
    if (diff.is_negative())
    {
        total_seconds--;
        fractional_seconds += 1'000'000'000;
    }
    _dst->set_seconds(total_seconds);
    _dst->set_nanos(fractional_seconds);
}

void proto_wrap(const boost::asio::ip::address& _src, Api::SourceIp* _dst)
{
    _dst->set_value(_src.to_string());
}

void proto_wrap(
        const LibLogger::LogEntryProperty& _src,
        google::protobuf::Map<std::string, Api::LogEntryProperties>* _dst)
{
    Api::LogEntryProperties props;
    props.mutable_values()->Reserve(_src.property_values.size());
    for (const auto& prop : _src.property_values)
    {
        const auto& dst_prop = props.add_values();
        dst_prop->set_value(prop.property_value.get());
        for (const auto& child : prop.children)
        {
            const auto& child_property_type = child.first;
            for (const auto& child_property_value : child.second)
            {
                (*dst_prop->mutable_children())[child_property_type.get()] = child_property_value.get();
            }
        }
    }
    if ((*_dst)[_src.property_type.get()].values().size() > 0) // deprecated case (input/output properties merge)
    {
        for (auto& prop : props.values())
        {
            const auto& dst_prop = ((*_dst)[_src.property_type.get()].add_values());
            dst_prop->set_value(prop.value());
            for (auto& child : prop.children())
            {
                (*dst_prop->mutable_children())[child.first] = child.second;
            }
        }
    }
    else {
        (*_dst)[_src.property_type.get()] = props;
    }
}

void proto_wrap(
        const LibLogger::ObjectReferences& _src,
        google::protobuf::Map<std::string, Api::ObjectReferenceValues>* _dst)
{
    Api::ObjectReferenceValues object_refs;
    object_refs.mutable_references()->Reserve(_src.object_reference_ids.size());
    for (const auto& id : _src.object_reference_ids)
    {
        object_refs.add_references()->set_value(id.get());
    }
    (*_dst)[_src.object_type.get()] = object_refs;
}

void proto_wrap(
        const std::map<LibLogger::LogEntry::ResultName, LibLogger::LogEntry::ResultCode>& _src,
        google::protobuf::Map<google::protobuf::string, google::protobuf::uint64>* _dst)
{
    for (const auto& src_item : _src)
    {
        (*_dst)[src_item.first.get()] = src_item.second.get();
    }
}

std::string to_Api_LogEntryId_value(const LibLogger::LogEntryIdent& _log_entry_ident)
{
    return std::to_string(_log_entry_ident.id.get()) + "#" +
            boost::posix_time::to_iso_string(_log_entry_ident.time_begin.get()) + "#" +
            _log_entry_ident.service_handle.get() + "#" +
            std::to_string(_log_entry_ident.monitoring_status.get());
}

std::string to_Api_SessionId_value(const LibLogger::SessionIdent& _session_ident)
{
    return std::to_string(_session_ident.id.get()) + "#" +
            boost::posix_time::to_iso_string(_session_ident.login_date.get());
}

} // namespace Fred::Logger::{anonymous}

void proto_wrap(const LibLogger::CreateLogEntryReply& _src, Api::CreateLogEntryReply* _dst)
{
    _dst->mutable_data()->mutable_log_entry_id()->set_value(to_Api_LogEntryId_value(_src.log_entry_ident));
}

void proto_wrap(const LibLogger::CreateSessionReply& _src, Api::CreateSessionReply* _dst)
{
    _dst->mutable_data()->mutable_session_id()->set_value(to_Api_SessionId_value(_src.session_ident));
}

void proto_wrap(const LibLogger::GetLogEntryTypesReply& _src, Api::GetLogEntryTypesReply* _dst)
{
    for (const auto& log_entry_type : _src.log_entry_types)
    {
        _dst->mutable_data()->add_types()->set_value(log_entry_type.get());
    }
}

void proto_wrap(const LibLogger::GetServicesReply& _src, Api::GetServicesReply* _dst)
{
    for (const auto& log_entry_service : _src.log_entry_services)
    {
        _dst->mutable_data()->add_services()->set_value(log_entry_service.get());
    }
}

void proto_wrap(const LibLogger::GetResultsReply& _src, Api::GetResultsReply* _dst)
{
    proto_wrap(_src.result_code_map, _dst->mutable_data()->mutable_result_code_map());
}

void proto_wrap(const LibLogger::GetObjectReferenceTypesReply& _src, Api::GetObjectReferenceTypesReply* _dst)
{
    for (const auto& object_reference_type : _src.object_reference_types)
    {
        _dst->mutable_data()->add_object_types()->set_value(object_reference_type.get());
    }
}

void proto_wrap(const LibLogger::SessionDoesNotExist&, Api::CloseSessionReply* _dst)
{
    _dst->mutable_exception()->mutable_session_does_not_exist();
}

void proto_wrap(const LibLogger::LogEntryDoesNotExist&, Api::CloseLogEntryReply* _dst)
{
    _dst->mutable_exception()->mutable_log_entry_does_not_exist();
}

void proto_wrap(const LibLogger::ListLogEntriesReply& _src, Api::ListLogEntriesReply* _dst)
{
    for (const auto& ident : _src.log_entry_ids)
    {
        _dst->mutable_data()->add_log_entry_ids()->set_value(to_Api_LogEntryId_value(ident));
    }
}

void proto_wrap(const LibLogger::ServiceNotFound&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_log_entry_service_not_found();
}

void proto_wrap(const LibLogger::InvalidTimestampFrom&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_invalid_timestamp_from();
}

void proto_wrap(const LibLogger::InvalidTimestampTo&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_invalid_timestamp_to();
}

void proto_wrap(const LibLogger::InvalidTimestampRange&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_invalid_timestamp_range();
}

void proto_wrap(const LibLogger::InvalidCountLimit&, Api::ListLogEntriesReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_invalid_count_limit();
}

void proto_wrap(const LibLogger::GetLogEntryInfoReply& _src, Api::GetLogEntryInfoReply* _dst)
{
    Api::LogEntryInfo* dst_info = _dst->mutable_data()->mutable_log_entry_info();
    const LibLogger::LogEntryInfo src_info = _src.log_entry_info;

    proto_wrap(src_info.time_begin.get(), dst_info->mutable_time_begin());

    dst_info->mutable_service()->set_value(src_info.service_name.get());
    if (src_info.time_end != boost::none && !src_info.time_end->get().is_special())
    {
        proto_wrap(src_info.time_end->get(), dst_info->mutable_time_end());
    }
    if (src_info.source_ip != boost::none)
    {
        proto_wrap(src_info.source_ip->get(), dst_info->mutable_source_ip());
    }
    if (src_info.log_entry_type != boost::none)
    {
        dst_info->mutable_log_entry_type()->set_value(src_info.log_entry_type->get());
    }
    if (src_info.result_code != boost::none)
    {
        dst_info->set_result_code(src_info.result_code->get());
    }
    if (src_info.result_name != boost::none)
    {
        dst_info->set_result_name(src_info.result_name->get());
    }
    if (src_info.raw_request != boost::none)
    {
        dst_info->mutable_raw_request()->set_value(src_info.raw_request->get());
    }
    if (src_info.raw_response != boost::none)
    {
        dst_info->mutable_raw_response()->set_value(src_info.raw_response->get());
    }
    if (src_info.username != boost::none)
    {
        dst_info->mutable_username()->set_value(src_info.username->get());
    }
    if (src_info.user_id != boost::none)
    {
        dst_info->mutable_user_id()->set_value(src_info.user_id->get());
    }
    for (const auto& props : src_info.input_properties)
    {
        proto_wrap(props, dst_info->mutable_input_properties());
        proto_wrap(props, dst_info->mutable_properties()); // deprecated
    }
    for (const auto& props : src_info.output_properties)
    {
        proto_wrap(props, dst_info->mutable_output_properties());
        proto_wrap(props, dst_info->mutable_properties()); // deprecated
    }
    for (const auto& refs : src_info.object_references)
    {
        proto_wrap(refs, dst_info->mutable_references());
    }
}

void proto_wrap(const LibLogger::LogEntryDoesNotExist&, Api::GetLogEntryInfoReply* _dst)
{
    _dst->clear_data();
    _dst->mutable_exception()->mutable_log_entry_does_not_exist();
}

} // namespace Fred::Logger
} // namespace Fred
